/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */


/* * * * * * * * * * * *\
 *    Panel-Docklet    *
 *        v1.01        *
 * Jodli  (pw3@gmx.at) *
\* * * * * * * * * * * */


/*
 * IDEAS:
 * 
 * (possible?) fill() - ignoriert minimized (Abfrage dafür suchen)
 * (possible?) window-preferences (arbeitsflaeche wechseln, vordergrund, ...)
 * (really?) title (on hover)
 * translations
 */


const St = imports.gi.St;
const Main = imports.ui.main;
const Shell = imports.gi.Shell;
const Lang = imports.lang;
const Monitor = Main.layoutManager.primaryMonitor;
const Tweener = imports.ui.tweener;
const DND = imports.ui.dnd;
const Mainloop = imports.mainloop;
const Gio = imports.gi.Gio;
const AppFavorites = imports.ui.appFavorites;
const AppDisplay = imports.ui.appDisplay;
const PopupMenu = imports.ui.popupMenu;
const ModalDialog = imports.ui.modalDialog;
const Clutter = imports.gi.Clutter;

const Gettext = imports.gettext.domain('gnome-shell-extensions');
const _ = Gettext.gettext;

//const DANEL_DOCK_SETTINGS_SCHEMA = 'org.gnome.shell.extensions.panel-docklet';

const MINIMIZE_WINDOW=0;
const NEW_WINDOW=1;
const CLOSE_WINDOW=2;

const DOCK_PADDING_BOTTOM = 3;
const DOCK_PADDING_LEFT = 3;
const DOCK_PADDING_RIGHT = 3;
const DOCK_WS_LABEL_FONT_SIZE = 10;
const WS_LABEL_WIDTH = 20;
const WS_LINES_SPACE_BETWEEN = 3;
const WS_LINES_SPACE_LEFT = 4;
const ICON_MINIMIZED_OPACITY = 80;
const ICON_COPY_OPACITY = 50;

let myPanelDocklet, extention_path;


//mimics gsettings-schema behaviour.
function SettingsContainer(extention_path) {
	this._init(extention_path);
}
SettingsContainer.prototype = {
	_init: function(extention_path) {
		this._connector = {};
		this._file = Gio.file_new_for_path(extention_path + "/settings.json");
		if(this._file.query_exists(null)) {
			[flag, data] = this._file.load_contents(null);
			
			this._conf = (flag) ? JSON.parse(data) : {};
			this._error = false;
		}
		else
			this._error = true;
	},
	
	get_boolean: function(k) {
		return this._conf[k] || false;
	},
	get_double: function(k) {
		return this._conf[k] || 0;
	},
	get_int: function(k) {
		return this._conf[k] || 0;
	},
	get_enum: function(k) {
		return this._conf[k] || 0;
	},
	
	set_boolean: function(k, v) {
		this._save_data(k, v);
	},
	set_double: function(k, v) {
		this._save_data(k, v);
	},
	set_int: function(k, v) {
		this._save_data(k, v);
	},
	set_enum: function(k, v) {
		this._save_data(k, v);
	},
	
	_save_data: function(k, v) {
		this._conf[k] = v;
		this.emit('changed::'+k);
		
		if(this._error)
			return;
		this._file.replace_contents(JSON.stringify(this._conf), null, false, 0, null);
	},
	
	connect: function(k, f) {
		this._connector[k] = f;
	},
	disconnect: function(k) {
		delete this._connector[k];
	},
	emit: function(k) {
		if(this._connector[k])
			this._connector[k](true);
	}
}

function panelDocklet() {
	this._init();
}
panelDocklet.prototype = {
	_init: function() {
		let settings =	new SettingsContainer(extention_path),
		//let settings = new Gio.Settings({ schema: DANEL_DOCK_SETTINGS_SCHEMA }),
			settingsListener = [];
		
		this._settings = settings;
		this._loadConfigs(false);
		
		
		this._dockHeight = this._DATA_default_height;
		
		this.actor = new St.BoxLayout({ name: 'panelDocklet',
				reactive: true,
				x: this._DATA_real_x,
				y: 0,
				width: this._DATA_default_width + this._DATA_padding_x_sum,
				height: this._dockHeight + DOCK_PADDING_BOTTOM,
				style_class: 'panelDocklet' });
		
		this._dragPointer = new St.Bin({style_class: 'moveTo', opacity:0});
		this.actor.add(this._dragPointer);
		
		Main.layoutManager.addChrome(this.actor);
		
		this._box = new workspaceBox(this);
		this._box._indexAllWS();
		
		this._isHidden = false;
		
		if((this._activateHiding = this._CONF_autohide))
			this._hideDock();
		else if(this._CONF_showAllWS)
			this._box.listWS();
		
		this._menuIsOpened = false;
		this._ID_menuTimeout = false;
		this._menuManager = new PopupMenu.PopupMenuManager(this);
		
		this._ID_leaveEvent = this.actor.connect('leave-event', Lang.bind(this, this._hideDock));
		this._ID_enterEvent = this.actor.connect('enter-event', Lang.bind(this, this._showDock));
		this._ID_clickEvent = this.actor.connect('button-release-event', Lang.bind(this, this._onMouseClick));
		
		this._ID_settingsListener = settingsListener;
		settingsListener.push(settings.connect('changed::autohide', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::hide-time', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::show-time', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::time-until-hide', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::middle-click-action', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::basic-icon-size', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::dock-x-right-from-center', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::dock-x-left', Lang.bind(this, this._loadConfigs)));
		settingsListener.push(settings.connect('changed::list-workspaces-on-open', Lang.bind(this, reload)));
		settingsListener.push(settings.connect('changed::show-workspace-number', Lang.bind(this, reload)));
		settingsListener.push(settings.connect('changed::bottom-hide-to', Lang.bind(this, this._loadConfigs)));
		
		if(this._settings.get_boolean('first-time'))
			preferences(this);
	},
	_loadConfigs: function(setting) {//setting: I dont use it, just to check, if function was called by connector
		this._CONF_autohide = this._settings.get_boolean('autohide');
		this._CONF_hideTime = this._settings.get_double('hide-time');
		this._CONF_showTime = this._settings.get_double('show-time');
		this._CONF_menuTimeout = this._settings.get_int('time-until-hide');
		this._CONF_hideToBottomPos = this._settings.get_int('bottom-hide-to');
		if(this._CONF_hideToBottomPos < 0)
			this._CONF_hideToBottomPos = 0;
		else if(this._CONF_hideToBottomPos > this._CONF_iconSize-1)
			this._CONF_hideToBottomPos = this._CONF_iconSize-1;
			
		this._CONF_middleClick = this._settings.get_enum('middle-click-action');
		
		this._CONF_iconSize = this._settings.get_int('basic-icon-size');
		if(this._CONF_iconSize < 10)
			this._CONF_iconSize = 10;
		
		this._CONF_dockX_right = this._settings.get_int('dock-x-right-from-center');
		this._CONF_dockX_left = this._settings.get_int('dock-x-left');
		
		this._CONF_showAllWS = this._settings.get_boolean('list-workspaces-on-open'); //if changed, needs reload
		this._CONF_showWSline = this._settings.get_boolean('show-workspace-number'); //if chenged, needs reload
		
		this._DATA_space_left = DOCK_PADDING_LEFT + (this._CONF_showWSline ? WS_LABEL_WIDTH : 0);
		this._DATA_padding_x_sum = this._DATA_space_left + DOCK_PADDING_RIGHT;
		this._DATA_default_width = Monitor.width/2 - this._CONF_dockX_left - this._CONF_dockX_right;
		this._DATA_default_height = this._CONF_iconSize;
		this._DATA_real_x = Monitor.x + this._CONF_dockX_left;
		this._DATA_max_icons = Math.floor(this._DATA_default_width / this._CONF_iconSize);
		
		if(setting) {
			this.actor.set_width(this._DATA_default_width + this._DATA_padding_x_sum);
			
			this.actor.set_x(this._DATA_real_x);
			
			let w = this._box._workspaces;
			let i = w.length;
			while(i--)
				w[i]._updateIconSize(true);
			
			if(this._box._WSareListed)
				this._box.unlistWS();
			
			if((this._activateHiding = this._CONF_autohide))
				this._hideDock();
			else
				this._showDock();
		}
	},
	
	_onMouseClick: function(actor, event) {
		/*if(event.get_button()==1) {
			let y = this._box._calcListY();
			if(event.y > y) {
				let wsI = (event.y-y) / (this._CONF_iconSize + WS_LINES_SPACE_BETWEEN);
				global.screen.get_workspace_by_index(wsI).activate(global.get_current_time());
			}
			return;
		}
		else */if(this._menuIsOpened || this._isHidden || event.get_button() != 3)
			return;
		this._activateHiding = false;
		this._menuIsOpened = true;
		
		if(!this._menu)
			this._menu = new DockletMenu(this);
		this._menu.popup();
	},
	
	_removeMenuTimeout: function() {
		if(this._ID_menuTimeout) {
			Mainloop.source_remove(this._ID_menuTimeout);
			this._ID_menuTimeout = false;
		}
	},
	_hideDock: function() {
		if(!this._activateHiding || this._ID_menuTimeout)
			return;
		
		this._isHidden=true;
		this._ID_menuTimeout = Mainloop.timeout_add(this._CONF_menuTimeout, Lang.bind(this, function() {
				if(this._CONF_menuTimeout && this._box._WSareListed)
					this._box.unlistWS();
				else {
					let s = this._dockHeight;
					Tweener.addTween(this.actor, {
							y: this._CONF_hideToBottomPos - s,
							time: this._CONF_hideTime,
							transition: 'easeOutQuad'
						});
				}
				this._ID_menuTimeout = false;
			}));
	},
	_showDock: function() {
		this._removeMenuTimeout();
		this._isHidden=false;
		
		let ws = this._box._workspaces;
		if(this._CONF_showAllWS && (ws.length>2 || (ws.length==2 && ws[1]._windows.length) || global.screen.get_active_workspace().index()))
			this._box.listWS();
		else
			Tweener.addTween(this.actor, {
					y: 0,
					time: this._CONF_showTime,
					transition: 'easeOutQuad'
				});
	},
	
	setDockHeight: function(s, yPos) {
		if(s && s != this._dockHeight) {
			this._dockHeight = s;
			if(this._CONF_showWSline && s<=this._CONF_iconSize)
				this._box._setWSlabelPadding(s);
		}
		else
			s = this._dockHeight;
		
		Tweener.addTween(this.actor, {
				height:DOCK_PADDING_BOTTOM + s,
				y: yPos || 0,
				time: yPos ? this._CONF_hideTime : this._CONF_showTime,
				transition: 'easeOutQuad'
			});
		
	},
	
	_destroy: function() {
		let l = this._ID_settingsListener,
			i = l.length;
		while(i--)
			this._settings.disconnect(l[i]);
			
		this.actor.disconnect(this._ID_leaveEvent);
		this.actor.disconnect(this._ID_enterEvent);
		this.actor.disconnect(this._ID_clickEvent);
		this._removeMenuTimeout();
		
		this._box._destroy();
		this.actor.destroy();
	}
}


function workspaceBox(dock) {
	this._init(dock);
}
workspaceBox.prototype = {
	_init: function(dock) {
		this._dock = dock;

		this._WSareListed = false;
		this._currentWindowFocus = false;
		this._workspaces = [];
		this._allWsWindows = [];
		
		if(dock._CONF_showWSline) {
			this._wsLabels = [];
			
			this._currentWSlabel = new St.Label({text:"1", x:0, y:0, width:WS_LABEL_WIDTH, style_class:"wsLabelMain"});
			dock.actor.add(this._currentWSlabel);
			this._currentWSlabel_paddingMinus = DOCK_WS_LABEL_FONT_SIZE;
			this._setWSlabelPadding();
		}
			
		if(dock._CONF_showAllWS)
			this._wsbg = new St.Bin({style_class: 'wsListBack', x:dock._DATA_space_left, width:dock._DATA_default_width});
		
		
		this._ID_newWorkspace = global.screen.connect('notify::n-workspaces', Lang.bind(this, this._newWS));
		this._ID_switchWorkspace = global.window_manager.connect('switch-workspace', Lang.bind(this, this._changeWS));
		
		this._ID_minimize = global.window_manager.connect('minimize', Lang.bind(this, this._onMinimize));
		this._ID_map = global.window_manager.connect('map', Lang.bind(this, this._onMap));
		
		let tracker = Shell.WindowTracker.get_default();
		this._ID_focusApp = tracker.connect('notify::focus-app', Lang.bind(this, this._onFocus));
	},
	_setWSlabelPadding: function(s) {
		this._currentWSlabel.set_style("padding: "+Math.round(((s || this._dock._CONF_iconSize) - this._currentWSlabel_paddingMinus) / 2)+"px 0;");
	},
	
	_indexAllWS: function() {
		let i = global.screen.n_workspaces;
		
		while(i--) {
			this._workspaces[i] = new windowBox(global.screen.get_workspace_by_index(i), this, this._dock);
		}
		if(this._dock._CONF_showWSline)
			this._currentWSlabel.text = (global.screen.get_active_workspace_index()+1).toString();
	},
	_newWS: function(ws, b) {
		//code-parts from: usr/share/gnome-shell/js/ui/main.js
		
		let oldNumWorkspaces = this._workspaces.length,
			newNumWorkspaces = global.screen.n_workspaces;
		
		if(newNumWorkspaces > oldNumWorkspaces) {
			// Assume workspaces are only added at the end
			for(let i = oldNumWorkspaces; i < newNumWorkspaces; ++i)
				this._workspaces[i] = new windowBox(global.screen.get_workspace_by_index(i), this, this._dock);
		}
		else if(newNumWorkspaces < oldNumWorkspaces){
			// Assume workspaces are only removed sequentially
			// (e.g. 2,3,4 - not 2,4,7)
			let removedNum = oldNumWorkspaces - newNumWorkspaces,
				removedIndex;
			for(let i=0; i < oldNumWorkspaces; ++i) {
				if(this._workspaces[i]._myWorkspace != global.screen.get_workspace_by_index(i)) {
					removedIndex = i;
					break;
				}
			}

			let lostWorkspaces = this._workspaces.splice(removedIndex, removedNum);
			lostWorkspaces.forEach(function(winBox) {
					winBox._destroy();
				});
		}
		
		//no idea why this is necessary.
		//But multiple app-windows on different workspaces wont be recognized properly without reconnect...
		let i = this._workspaces.length;
		while(i--) {
			this._workspaces[i]._removeListener();
			this._workspaces[i]._setListener();
		}
		
		if(this._WSareListed) {
			this.unlistWS();
			this.listWS();
		}
		
		return false; //blind copy
	},
	_changeWS: function(wm, from, to, direction) {
		let movesUp = (from>to) ? false : true;
		
		if(to >= this._workspaces.length)
			this._workspaces[to] = new windowBox(global.screen.get_workspace_by_index(i), this, this._dock);
		
		this._workspaces[from]._removeAllIcons(movesUp);
		this._workspaces[to].activateIcons(!movesUp);
		
		if(this._WSareListed) {
			let confIconSize = this._dock._CONF_iconSize,
				start_sum = this._calcListY();
			this._workspaces[from].activateIcons(!movesUp, start_sum + (confIconSize + WS_LINES_SPACE_BETWEEN) * from);
			this._workspaces[to].activateCopies(!movesUp, start_sum + (confIconSize + WS_LINES_SPACE_BETWEEN) * to);
		}
		
		
		if(!this._WSareListed)
			this._workspaces[to]._updateHeight();
		
		if(this._dock._CONF_showWSline)
			this._currentWSlabel.text = (to+1).toString();
	},
	_calcListY: function() {
		let confIconSize = this._dock._CONF_iconSize;
		return confIconSize + Math.round(confIconSize/3);
	},
	listWS: function() {
		if(this._WSareListed)
			return;
		this._WSareListed = true;
		
		let ws = this._workspaces,
			cI = global.screen.get_active_workspace_index(),
			confIconSize = this._dock._CONF_iconSize,
			actor = this._dock.actor,
		
			wsLabels = this._wsLabels,
			start_sum = this._calcListY(),
			sum = start_sum,
			item;
		
		for(let i=0, max = ws.length; i<max; ++i) {
			if(i != cI)
				ws[i].activateIcons(true, sum);
			else
				ws[i].activateCopies(true, sum);
			
			if(this._dock._CONF_showWSline) {
				item = new St.Label({text:(i+1)+":", x:WS_LINES_SPACE_LEFT, y:sum + confIconSize/4});
				actor.add(item);
				wsLabels.push(item);
			}
			sum += confIconSize + WS_LINES_SPACE_BETWEEN;
		}
		
		item = this._wsbg;
		actor.add(item);
		item.set_y(start_sum);
		item.set_height(sum-start_sum);
		item.lower_bottom();
		
		this._dock.setDockHeight(sum);
	},
	unlistWS: function(cI) {
		if(!this._WSareListed)
			return;
		
		cI = (cI!=undefined) ? cI : global.screen.get_active_workspace_index(); //0==false
		let ws = this._workspaces,
			wsI = ws.length;
		
		//if new WS has been created or removed, ws.length and labels.length can be different
		if(this._dock._CONF_showWSline) {
			let labels = this._wsLabels,
				laI = labels.length;
			
			while(laI--) {
				labels[laI].destroy();
			}
			this._wsLabels = [];
		}
		this._dock.actor.remove_actor(this._wsbg);
		
		
		ws[cI]._updateHeight();
		while(wsI--) {
			if(wsI != cI)
				ws[wsI]._removeAllIcons(true, true);
		}
		
		
		ws[cI].removeAllCopies();
		this._WSareListed = false;
	},
	
	_onFocus: function() {
		if(this._currentWindowFocus) {
			this._currentWindowFocus._myWindowClass._checkFor_allWSWindow(this._currentWindowFocus);
			this._currentWindowFocus.removeFocus();
		}
		
		let f = global.display.focus_window, f_icon;
		if(f && (f_icon = f._PANEL_DOCKLET_icon)) {
			//we dont have to worry about all-workspace-windows. Just focus the first icon
			//and it will focus its doubles (including current workspace) too
			f_icon.showFocused();
			f_icon._myWindowClass._checkFor_allWSWindow(f_icon);
			
			this._currentWindowFocus = f_icon;
		}
		else
			this._currentWindowFocus = false;
	},
	_onMinimize: function(shellwm, actor) {
		actor.get_meta_window()._PANEL_DOCKLET_icon.showMinimized();
	},
    _onMap: function(shellwm, actor) {
		let a;
		if((a=actor.get_meta_window()._PANEL_DOCKLET_icon))
			a.showMaped();
	},
	
	_addAllWsWindow: function(app, win, notThisWS) {
		if(!win._PANEL_DOCKLET_addedToAllWS) {
			this._allWsWindows.push([app, win]);
			
			let ws = this._workspaces,
				wsI= ws.length;
			if(notThisWS) {
				while(wsI--) {
					if(ws[wsI]._myWorkspace != notThisWS)
						ws[wsI].addWindow(app, win);
				}
			}
			else {
				while(wsI--)
					ws[wsI].addWindow(app, win);
			}
			win._PANEL_DOCKLET_addedToAllWS = true;
			return true;
		}
		return false
	},
	_removeAllWsWindow: function(win, notAllWS) {
		let a = this._allWsWindows,
			i = a.length,
			
			ws = this._workspaces,
			wsI= ws.length;
		if(notAllWS) {
			let notIndex = global.screen.get_active_workspace_index(),
				icon = win._PANEL_DOCKLET_icon,
			
				lasIcon = (icon._myDoubles[notIndex] ? icon._myDoubles[notIndex] : icon);
			lasIcon._myDoubles = [];
			win._PANEL_DOCKLET_icon = lasIcon;
			
			
			while(wsI--) {
				if(wsI != notIndex)
					ws[wsI].removeWindow(win);
			}
		}
		else {
			while(wsI--)
				ws[wsI].removeWindow(win);
		}
		delete win._PANEL_DOCKLET_addedToAllWS;
		
		
		while(i--) {
			if(a[i][1] != win)
				continue;
			
			this._allWsWindows.splice(i, 1);
			return true;
			
		}
		return false;
	},
	
	_destroy: function() {
		global.screen.disconnect(this._ID_newWorkspace);
		global.window_manager.disconnect(this._ID_switchWorkspace);
		global.window_manager.disconnect(this._ID_minimize);
		global.window_manager.disconnect(this._ID_map);
		Shell.WindowTracker.get_default().disconnect(this._ID_focusApp);
		
		let i = this._workspaces.length;
		while(i--) {
			this._workspaces[i]._destroy();
		}
	}
}


function windowBox(ws, wsBox, dock) {
	this._init(ws, wsBox, dock);
}
windowBox.prototype = {
	_init: function(ws, wsBox, dock) {
		this._dock = dock;
		this._myWorkspace = ws;
		this._wsBox = wsBox;
		this._windows = [];
		this._isListedTo = 0; //0==false
		this._copiesListedTo = 0;
		
		this._iconSize = this._dock._CONF_iconSize;
		
		this._ID_windowAdded = false;
		this._ID_windowRemoved = false;
		
		
		this._indexAllWindows();
		
		this._setListener();
	},
	_indexAllWindows: function() {
		let realWindows = this._myWorkspace.list_windows(),
			i = realWindows.length;
		
		while(i--) {
			this._onWindowAdded(this._myWorkspace, realWindows[i]);
			
			if(realWindows[i].has_focus()) 
				this._wsBox._currentWindowFocus = this._windows[this._windows.length-1];//last element is currently added window
		}
	},
	_setListener: function() {
		let ws = this._myWorkspace;
		this._ID_windowAdded = ws.connect('window-added', Lang.bind(this, this._onWindowAdded));
		this._ID_windowRemoved = ws.connect('window-removed', Lang.bind(this, this._onWindowRemoved));
	},
	_removeListener: function() {
		let ws = this._myWorkspace;
		
		if(this._ID_windowAdded) {
			ws.disconnect(this._ID_windowAdded);
			this._ID_windowAdded = false;
		}
		if(this._ID_windowRemoved) {
			ws.disconnect(this._ID_windowRemoved);
			this._ID_windowRemoved = false;
		}
	},
	
	_onWindowAdded: function(metaWorkspace, metaWindow) {
		let tracker = Shell.WindowTracker.get_default(),
			app = tracker.get_window_app(metaWindow);
		
		if (app && tracker.is_window_interesting(metaWindow)) {
			if(!metaWindow.is_on_all_workspaces() || !this._wsBox._addAllWsWindow(app, metaWindow))
				this.addWindow(app, metaWindow);
		}
	},
	addWindow: function(app, win) {
		this._addIcon(app, win);
		this._updateIconSize();
	},
	_onWindowRemoved: function(metaWorkspace, metaWindow) {
		if(!metaWindow._PANEL_DOCKLET_addedToAllWS || !this._wsBox._removeAllWsWindow(metaWindow))
			this.removeWindow(metaWindow);
	},
	removeWindow: function(win) {
		let i = this._windows.length;
		while(i--) {
			if (this._windows[i]._myWindow == win) {
				this.deleteIcon(i);
				return;
			}
		}
	},
	
	_checkFor_allWSWindow: function(win) {
		let realwin = win._myWindow;
		if(realwin.is_on_all_workspaces() && !realwin._PANEL_DOCKLET_addedToAllWS)
			this._wsBox._addAllWsWindow(win._myApp, realwin, this._myWorkspace);
		
		else if(realwin._PANEL_DOCKLET_addedToAllWS && !realwin.is_on_all_workspaces())
			this._wsBox._removeAllWsWindow(realwin, true);
	},
	
	_addIcon: function(app, current_window) {
		let item = new windowIconButton(
				this._dock,
				app,
				current_window,
				this,
				this._windows.length, //index-count is current index
				((this._isListedTo || this._myWorkspace.index() == global.screen.get_active_workspace_index()) ? this._iconSize :false),
				this._isListedTo
			);
		this._windows.push(item);
		
		if(current_window._PANEL_DOCKLET_icon && current_window._PANEL_DOCKLET_icon != item) {
			//global.log(current_window._PANEL_DOCKLET_icon._myWindowClass._myWorkspace.index());
			current_window._PANEL_DOCKLET_icon._myDoubles[this._myWorkspace.index()] = item;
		}
		else
			current_window._PANEL_DOCKLET_icon = item;
		
		if(this._wsBox._WSareListed && this._myWorkspace.index() == global.screen.get_active_workspace_index())
			item.createCopy(this._iconSize, false, this._wsBox._calcListY() + this._myWorkspace.index() * (this._dock._CONF_iconSize + WS_LINES_SPACE_BETWEEN));
	},
	activateIcons: function(movesUp, yPos) {
		this._isListedTo = yPos || 0;
		
		let i = this._windows.length;
		while(i--) {
			this._windows[i].createIcon(this._iconSize, movesUp, yPos);
		}
	},
	activateCopies: function(movesUp, yPos) {
		this._copiesListedTo = yPos || 0;
		
		let i = this._windows.length,
			win;
		while(i--) {
			win = this._windows[i];
			this._checkFor_allWSWindow(win);
			
			this._windows[i].createCopy(this._iconSize, movesUp, yPos);
		}
	},
	removeAllCopies: function(movesUp) {
		this._windows.forEach(function(i) {
				i.removeCopies();
			});
	},
	deleteIcon: function(i) {
		let win = this._windows[i],
			realwin = win._myWindow;
		
		win._destroy();
		if(!realwin._PANEL_DOCKLET_addedToAllWS)
			realwin._PANEL_DOCKLET_icon = false;
		
		this._windows.splice(i, 1);
		this._updateIconSize();
	},
	_removeAllIcons: function(movesUp, noAnimation) {
		if(this._isListedTo)
			this._isListedTo = 0;
		
		let i = this._windows.length,
			win;
		
		while(i--) {
			win = this._windows[i];
			this._checkFor_allWSWindow(win);
			
			this._windows[i].deleteIcon(movesUp, true, false, noAnimation);
		}
	},
	_updateIconSize: function(update) {
		let l = this._windows.length;
		if(l >= this._dock._DATA_max_icons || update) {
			let s = (l <= this._dock._DATA_max_icons) ? this._dock._CONF_iconSize :(this._dock._DATA_default_width) / l,
				i = this._windows.length;
			
			this._iconSize = s;
			
			while(i--) {
				this._windows[i]._changeSize(s, update);
			}
			
			if(!this._wsBox._WSareListed)
				this._updateHeight(s);
		}
	},
	_updateHeight: function(s) {
		s = s || this._iconSize;
		let y = ((s<this._dock._CONF_hideToBottomPos || !this._dock._activateHiding) ? 0 : this._dock._CONF_hideToBottomPos - s);
		this._dock.setDockHeight(s, y);
	},
	
	_destroy: function() {
		let i = this._windows.length;
		while(i--) {
			this._windows[i]._destroy();
		}
		this._removeListener();
	}
}


function windowIconButton(dock, app, metaWindow, windowClass, i, iconSize, ypos) {
	this._init(dock, app, metaWindow, windowClass, i, iconSize, ypos);
}
windowIconButton.prototype = {
	_init: function(dock, app, window, windowClass, i, iconSize, ypos) {
		this._dock = dock;
		this._myWindow = window;
		this._myWindowClass = windowClass;
		this._myApp = app;
		this._myIndex = i;
		this._myDoubles = []; //link icons to other workspaces for same window
		this._myCopies = [];
		this._isMinimized=false;
		
		this._menuManager = new PopupMenu.PopupMenuManager(this);
		
		if(iconSize)//change of iconSize can only be calculated in outer Class, so remembering wouldnt make sense here
			this.createIcon(iconSize, false, ypos);
		else
			this.actor = null;
	},
	_calcX: function(iconSize) {
		return this._dock._DATA_space_left + this._myIndex*iconSize;
	},
	
	createCopy: function(iconSize, movesUp, newY) {
		this.createIcon(iconSize, movesUp, newY, true);
	},
	createIcon: function(iconSize, movesUp, newY, createAsCopy) {
		if(this.actor && !createAsCopy)
			this.deleteIcon(!movesUp, true);
		
		newY = newY || 0;
		
		let size = Math.floor(iconSize),
		
			item = new St.Button({ style_class: 'windowIconButton',
//				button_mask: St.ButtonMask.ONE | St.ButtonMask.TWO, //blind copy...
				reactive: true,
				can_focus: false,
				width: size,
				height: size,
				x: this._calcX(iconSize),
				y: newY + (movesUp ? -iconSize : iconSize),
				opacity: 0,
				x_fill: true,
				y_fill: true
			});
			
		item._delegate = this;
		let dragable = DND.makeDraggable(item),
			event_click = item.connect('button-release-event', Lang.bind(this, this._onMouseClick)),
			icon = this._myApp.create_icon_texture(size);
		
		dragable.connect('drag-begin', Lang.bind(this, this._startDrag));
		dragable.connect('drag-cancelled', Lang.bind(this, this._endDrag));
		dragable.connect('drag-end', Lang.bind(this, this._endDrag));
		item.set_child(icon);
		//item.set_tooltip_text(this._myWindow.title); //strange behaviour and warnings....
		
		
		this._dock.actor.add(item);
		
		Tweener.addTween(item, {
				opacity: createAsCopy ? ICON_COPY_OPACITY : 255,
				y:newY,
				time: 0.5,
				transition: 'easeOutExpo'
			});
		if(!createAsCopy) {
			this.actor = item;
			this._icon = icon;
			this._draggable = dragable;
			this._ID_buttonPressEvent = event_click;
		}
		else
			this._myCopies.push([item, icon, event_click]);
		
		this._showWindowStats();
	},
	removeCopies: function() {
		let i = this._myCopies.length;
		
		while(i--) {
			this.deleteIcon(true, true, this._myCopies[i]);
		}
		this._myCopies = [];
	},
	deleteIcon: function(movesUp, dontMoveOthers, copy, noAnimation) {
		let item, event_click;
		
		if(!dontMoveOthers)
			this._moveOthers();
		
		if(!copy) {
			item = this.actor;
			event_click = this._ID_buttonPressEvent;
			
			this.actor = null;
			this.removeCopies();
		}
		else {
			item = copy[0];
			event_click = copy[2];
		}
		
		if(!item)
			return;
		
		item.disconnect(event_click);
		
		if(!noAnimation) {
			Tweener.addTween(item, {
					opacity: 0,
					y: item.y + (movesUp ? -item.height : item.height),
					time: 0.6,
					transition: 'easeOutQuart',
					onComplete: function() {
						this.destroy();
					}
				});
		}
		else
			item.destroy();
	},
	
	_onMouseClick: function(actor, event) {
		let button = event.get_button();
		switch(button) {
			case 1://left
				if(this._myWindow.has_focus())
					this.doMinimize();
				else
					this.doMap();
			break;
			case 2://middle
				switch(this._dock._CONF_middleClick) {
					case MINIMIZE_WINDOW:
						this.doMinimize();
					break;
					case NEW_WINDOW:
						this.doNew();
					break;
					case CLOSE_WINDOW:
						this.doClose();
					break;
				}
			break;
			case 3://right
				if(this._dock._menuIsOpened || this._dock._isHidden)
					return;
				this._dock._activateHiding = false;
				this._dock._menuIsOpened = true;
				
				if(this._menu)
					this._menu._destroy();
				this._menu = new IconMenu(this);
				this._menu.popup();
		}
	},
	
	_startDrag: function() {
		if(this._menu)
			this._menu._destroy();
		this.actor.opacity = 60;
		
		this._dragMonitor = {
				dragMotion: Lang.bind(this, this._onDragMotion)
			};
		DND.addDragMonitor(this._dragMonitor);
		this._dragState = true;
	},
	_endDrag: function() {
		if(!this._dragState)
			return;
		this._dragState = false;
		
		DND.removeDragMonitor(this._dragMonitor);
		this._dock._dragPointer.opacity = 0; //actor has to stay because of drag-Animation...
		
		if(!this.actor) //happens, when WS is changed
			return;
		this.actor.opacity = 255;
		this._showWindowStats();
	},
	_onDragMotion: function(drag) {
		let _this = this,
			current_wsI = this._myWindowClass._myWorkspace.index(),
			pointer = this._dock._dragPointer,
			x = drag.x - this._dock.actor.x,
			y = drag.y - this._dock.actor.y,
		
			startY = this._dock._CONF_iconSize + Math.round(this._dock._CONF_iconSize/3),
		
			wsI = (y < startY) ?
				global.screen.get_active_workspace().index() :
				Math.floor((y - startY) / (this._dock._CONF_iconSize + WS_LINES_SPACE_BETWEEN)),
		
			ws = this._dock._box._workspaces,
		
			winBox = ws[(wsI >=ws.length) ? ws.length-1 : wsI],
			winI = Math.floor((x-this._dock._DATA_space_left) / winBox._iconSize),
		
			winLe = winBox._windows.length;
		
		if(winI < 0)
			winI = 0;
		else if(winI >= winLe)
			winI = (winLe >= this._dock._DATA_max_icons || wsI == current_wsI) ? winLe-1 : winLe;
		
		if(wsI == current_wsI && winI != this._myIndex) {
			this._dragMonitor.dragDrop = function() {
					_this._swapTo(winI);
					
					return DND.DragDropResult.CONTINUE;
				};
		}
		else if(wsI != current_wsI && !this._myWindow.is_on_all_workspaces()) {
			this._dragMonitor.dragDrop = function() {
					let wins = winBox._windows,
						l = wins.length;
					
					_this._myWindow.change_workspace_by_index(
							wsI,
							false, // don't create workspace
							global.get_current_time()
						);
					
					//to make sure, window already exists and icon is resized
					//FIXME: bad solution...
					if(winI < l)
						this._ID_menuTimeout = Mainloop.timeout_add(400, Lang.bind(this, function() {
								wins[l]._swapTo(winI);
							}));
					
					return DND.DragDropResult.CONTINUE;
				};
			
		}
		else {
			this._dragMonitor.dragDrop = false;
			
			pointer.set_position(this.actor.x, this.actor.y);
			pointer.set_size(this.actor.width, this.actor.height);
			
			return DND.DragMotionResult.NO_DROP;
		}
		
		
		if(this._dragMonitor.dragDrop) {
			let iconSize = winBox._iconSize,
				px = winI * iconSize + this._dock._DATA_space_left,
				py = winBox._isListedTo;
			
			pointer.set_position(px, py);
			pointer.set_size(iconSize, iconSize);
			
			return DND.DragMotionResult.MOVE_DROP;
		}
		else
			return DND.DragMotionResult.CONTINUE;
	},
	getDragActor: function() {
        let item = this._myApp.create_icon_texture(this._icon.width);
        item.opacity = 180;
        return item;
    },
    getDragActorSource: function() {
		let item = this._dock._dragPointer;
		item.set_position(this.actor.x, this.actor.y);
        item.opacity = 255;
		item.raise_top();
		return item;
    },
	
	_moveOthers: function() {
		let wins = this._myWindowClass._windows,
			i = wins.length;
		while(--i > this._myIndex) { //I guess thats faster than for()... I guess...
			wins[i]._moveLeft();
		}
	},
	_moveLeft: function() {
		--this._myIndex;
		this._movingTween();
	},
	_moveRight: function() {
		++this._myIndex;
		this._movingTween();
	},
	_movingTween: function() {
		if(!this.actor)
			return;
		let copies = this._myCopies;
		Tweener.addTween(this.actor, {
				x: this._calcX(this.actor.width),
				time: 0.5,
				transition: 'easeInQuart',
				onUpdate:(copies.length ? function() {
						let i = copies.length;
						while(i--)
							copies[i][0].x = this.x;
					} : false)
			});
	},
	_swapTo: function(toI) {
		//FIXME: this function is cruel...
		let w = this._myWindowClass._windows,
			i = this._myIndex,
			t = w[i];
		
		
		if(toI < i) //left
			while(--i >= toI) {
				w[i]._moveRight();
				w[i+1] = w[i];
			}
		else //right
			while(++i <= toI) {
				w[i]._moveLeft();
				w[i-1] = w[i];
			}
		
		this._myIndex = toI;
		w[toI] = t; //now: i==toI
		
		if(!this.actor)
			return;
		
		Tweener.addTween(this.actor, {
				x: this._calcX(this.actor.width),
				time: 0.5,
				transition: 'easeInQuart'
			});
		
		if(this._myCopies.length) {
			let xpos = this._calcX(this.actor.width);
			
			this._myCopies.forEach(function(c) {
				Tweener.addTween(c[0], {
						x: xpos,
						time: 0.5,
						transition: 'easeInQuart'
					});
				});
		}
	},
	_swapIcons: function(moveRight) {
		let w = this._myWindowClass._windows,
			mI = this._myIndex,
			i1, i2;
		
		if(moveRight && mI+1 < w.length) {//right
			i1 = mI;
			i2 = mI+1;
		}
		else if(!moveRight && mI) {//left (mI > 0)
			i1 = mI-1;
			i2 = mI;
		}
		else
			return;
		
		
		let t = w[i1];
		w[i1] = w[i2];
		w[i2] = t;
		
		w[i1]._moveLeft();
		w[i2]._moveRight();
	},
	_changeSize: function(s, update) {
		if(!this.actor)
			return;
		let size = Math.floor(s),
			copies = this._myCopies;
		Tweener.addTween(this.actor, {
				x: this._calcX(s),
				width: size,
				height: size,
				time: 0.3,
				transition: 'easeInOutCubic',
				onUpdate:(copies.length ? function() {
						let w = this.width,
							x = this.x,
							i = copies.length,
							c;
						while(i--) {
							c = copies[i][0];
							c.width = w;
							c.height = w;
							c.x = x;
						}
					} : false)
			});
		if(update) {
			this._icon.destroy();
			this._icon = this._myApp.create_icon_texture(size);
			this.actor.set_child(this._icon);
		}
	},
	
    doMinimize: function() {
		this._myWindow.minimize(global.get_current_time());
		this.showMinimized();
	},
	doMap: function() {
		let win = this._myWindow;
		if(win.get_workspace() != global.screen.get_active_workspace())
			win.get_workspace().activate(global.get_current_time());
		win.activate(global.get_current_time());
		this.showMaped();
	},
	doNew: function() {
		this._myApp.open_new_window(global.screen.get_active_workspace().index());
	},
	doClose: function() {
		this._myWindow.delete(global.get_current_time());
	},
	
	_showWindowStats: function() {
		if(this._myWindow.has_focus() && this._myWindow.get_workspace().index() == global.screen.get_active_workspace_index())
			this.showFocused();
		if(this._isMinimized)
			this.showMinimized();
	},
	showFocused: function() {
		if(this.actor)
			this.actor.add_style_pseudo_class('focused');
		
		if(this._myDoubles.length)
			this._myDoubles.forEach(function(d) {
					if(d.actor)
						d.actor.add_style_pseudo_class('focused');
					if(d._myCopies.length)//all-ws-windows can have their icon on inactive ws
						d._myCopies.forEach(function(c) {c[0].add_style_pseudo_class('focused');});
				});
		
		if(this._myCopies.length)
			this._myCopies.forEach(function(c) {c[0].add_style_pseudo_class('focused');});
	},
	removeFocus: function() {
		if(this.actor)
			this.actor.remove_style_pseudo_class('focused');
		
		if(this._myDoubles.length)
			this._myDoubles.forEach(function(d) {
					if(d.actor)
						d.actor.remove_style_pseudo_class('focused');
					if(d._myCopies.length)//all-ws-windows can have their icon on inactive ws
						d._myCopies.forEach(function(c) {c[0].remove_style_pseudo_class('focused');});
				});
			
		if(this._myCopies.length)
			this._myCopies.forEach(function(c) {c[0].remove_style_pseudo_class('focused');});
	},
	showMinimized: function() {
		this._isMinimized=true;
		this._icon.opacity = ICON_MINIMIZED_OPACITY;
		
		if(this._myDoubles.length) {
			this._myDoubles.forEach(function(d) {
					d._isMinimized=true;
					d._icon.opacity = ICON_MINIMIZED_OPACITY;
				});
		}
		
		if(this._myCopies.length)
			this._myCopies.forEach(function(c) {c[1].opacity = ICON_MINIMIZED_OPACITY;});
	},
	showMaped: function() {
		this._isMinimized=false;
		this._icon.opacity = 255;
		
		
		if(this._myDoubles.length)
			this._myDoubles.forEach(function(d) {
					d._isMinimized=false;
					d._icon.opacity = 255;
				});
		
		if(this._myCopies.length)
			this._myCopies.forEach(function(c) {c[1].opacity = 255;});
	},
	
	_destroy: function() {
		this.deleteIcon();
		this.removeCopies();
	}
};

function DockletMenu(source) {
    this._init(source);
}
DockletMenu.prototype = {
	__proto__: AppDisplay.AppIconMenu.prototype,
	_init: function(source) {
		this._source = source;
		PopupMenu.PopupMenu.prototype._init.call(this, source.actor, St.Align.MIDDLE, St.Side.LEFT, 0);

        this.actor.add_style_class_name('panelDocklet_popup');

		this._ID_activate = this.connect('activate', Lang.bind(this, this._onActivate));
		
		this._ID_openStateChanged = this.connect('open-state-changed', Lang.bind(this, function (menu, isPoppedUp) {
			if (!isPoppedUp) {
				if((this._source._activateHiding = this._source._CONF_autohide)) //Variable is changed here!
					this._source._hideDock();
				
				this._source._menuIsOpened = false;
			}
		}));
		
		Main.layoutManager.addChrome(this.actor);
		
		source._menuManager.addMenu(this, true);
    },
	_redisplay: function() {
        this.removeAll();
		this._settingsItem = this._appendMenuItem(_("Settings"));
		this._appendSeparator();
		
		let launchers = global.settings.get_strv(AppFavorites.getAppFavorites().FAVORITE_APPS_KEY),
			i = launchers.length,
			item, icon;
		while(i--) {
			let app = Shell.AppSystem.get_default().lookup_app(launchers[i]);

            if(!app)
				continue;
				
			item = new PopupMenu.PopupMenuItem(app.get_name());
			item._app = app
			this.addMenuItem(item);
			
			icon = app.create_icon_texture(24);
			item.addActor(icon, {align: St.Align.END});
		}
	},
	_onActivate: function(actor, child) {
		if(child == this._settingsItem)
			preferences(this._source, true);
		else if(child._app) {
			let current_ws = global.screen.get_active_workspace().index();
			child._app.open_new_window(current_ws);
		}
	},
	_destroy: function() {
		if(this.isOpen)
			this._source._menuIsOpened = false;
		this.actor.destroy();
	}
}

function IconMenu(source) {
    this._init(source);
}
IconMenu.prototype = {
	__proto__: AppDisplay.AppIconMenu.prototype,

	_init: function(source) {
		this._source = source;
		this._dock = source._dock;
		PopupMenu.PopupMenu.prototype._init.call(this, source.actor, St.Align.MIDDLE, St.Side.TOP, 0);

        this.actor.add_style_class_name('panelDocklet_popup');

		this._ID_activate = this.connect('activate', Lang.bind(this, this._onActivate));

		this._ID_destroy = source.actor.connect('destroy', Lang.bind(this, this._destroy));
		
		this._ID_openStateChanged = this.connect('open-state-changed', Lang.bind(this, function (menu, isPoppedUp) {
			if (!isPoppedUp) {
				this._dock._menuIsOpened = false;
				if((this._dock._activateHiding = this._dock._CONF_autohide)) //Variable is changed here!
					this._dock._hideDock();
			}
		}));
		
		Main.layoutManager.addChrome(this.actor);
		
		source._menuManager.addMenu(this, true);
    },
	_removeListener: function() {
		if(this._ID_clickedLeft && this._Item_moveLeft)
			this._Item_moveLeft.disconnect(this._ID_clickedLeft);
		if(this._ID_clickedRight && this._Item_moveRight)
			this._Item_moveRight.disconnect(this._ID_clickedRight);
		
		if(this._ID_clickedUp && this._Item_moveUp)
			this._Item_moveUp.disconnect(this._ID_clickedUp);
		if(this._ID_clickedDown && this._Item_moveDown)
			this._Item_moveDown.disconnect(this._ID_clickedDown);
		
		if(this._ID_openStateChanged)
			this.disconnect(this._ID_openStateChanged);
		if(this._ID_activate)
			this.disconnect(this._ID_activate);
		if(this._ID_destroy && this._source.actor)
			this._source.actor.disconnect(this._ID_destroy);
		
		if(this._ID_quitApp && this._item_quitApp)
			this._item_quitApp.disconnect(this._ID_quitApp);
		if(this._ID_toggleFav && this._Item_toggleFav)
			this._Item_toggleFav.disconnect(this._ID_toggleFav);
			
		this._ID_clickedLeft = false;
		this._ID_clickedRight = false;
		this._ID_clickedUp = false;
		this._ID_clickedDown = false;
		this._ID_openStateChanged = false;
		this._ID_activate = false;
		this._ID_destroy = false;
		this._ID_quitApp = false;
		this._ID_toggleFav = false;
	},
	_redisplay: function() {
		let source = this._source,
			section = new St.BoxLayout({vertical: false}),
			index = source._myIndex;
		
		//left
		if(index) {//index > 0
			this._item_moveLeft = new St.Button({ style_class: 'swapWindowIcon left' });
			section.add(this._item_moveLeft, {expand: false});
			this._ID_clickedLeft = this._item_moveLeft.connect('clicked', Lang.bind(this, this._onActivate));
		}
		//right
		if(index+1 < source._myWindowClass._windows.length) {
			this._item_moveRight = new St.Button({ style_class: 'swapWindowIcon right' });
			section.add(this._item_moveRight, {expand: false, x_align: St.Align.MIDDLE});
			this._ID_clickedRight = this._item_moveRight.connect('clicked', Lang.bind(this, this._onActivate));
		}
		
		//middle
		let label = new St.Label( {style_class: 'labelLine'});
        section.add(label, { expand: true, x_fill: true, x_align: St.Align.MIDDLE });
		
		if(!this._source._myWindow.is_on_all_workspaces()) {
			//up
			if(this._source._myWindowClass._myWorkspace.index()) {//index > 0
				this._item_moveUp = new St.Button({ style_class: 'swapWindowIcon up' });
				section.add(this._item_moveUp, {expand: false});
				this._ID_clickedUp = this._item_moveUp.connect('clicked', Lang.bind(this, this._onActivate));
			}
			//down
			this._item_moveDown = new St.Button({ style_class: 'swapWindowIcon down' });
			section.add(this._item_moveDown, {expand: false});
			this._ID_clickedDown = this._item_moveDown.connect('clicked', Lang.bind(this, this._onActivate));
		}
		
		this.addActor(section);
		
		
		//close
		this._item_closeWin = new PopupMenu.PopupMenuItem(_("Close Window"), {style_class: 'specialOne'});
		this.addMenuItem(this._item_closeWin);
		
		//map/ minimized
		if(source._isMinimized) {
			this._item_mapWin = new PopupMenu.PopupMenuItem(_("Restore"), {style_class: 'specialOne'});
			this.addMenuItem(this._item_mapWin);
		}
		else {
			this._item_minimizeWin = new PopupMenu.PopupMenuItem(_("Minimize"), {style_class: 'specialOne'});
			this.addMenuItem(this._item_minimizeWin);
		}
		
		//new
		this._item_newWin = new PopupMenu.PopupMenuItem(_("New Window"), {style_class: 'specialOne'});
		this.addMenuItem(this._item_newWin);
		
		
		let subsection = new PopupMenu.PopupSubMenuMenuItem(_("Application"));
		
		this._item_quitApp = new PopupMenu.PopupMenuItem(_("Quit Application"), {style_class: 'specialOne'});
		subsection.menu.addMenuItem(this._item_quitApp);
		
		this._Item_toggleFav = new PopupMenu.PopupMenuItem(_((AppFavorites.getAppFavorites().isFavorite(this._source._myApp.get_id()) ? 
				"Remove from Favorites" :
				"Add to Favorites"
			)), {style_class: 'specialOne'});
		subsection.menu.addMenuItem(this._Item_toggleFav);
		
		this.addMenuItem(subsection);
		
		this._ID_quitApp = this._item_quitApp.connect('activate', Lang.bind(this, this._onActivate));
		this._ID_toggleFav = this._Item_toggleFav.connect('activate', Lang.bind(this, this._onActivate));
		
		//windows
		let windows = source._myApp.get_windows();
		if (windows.length) {
			this._appendSeparator();
			
			//Asuming: fewer windows -> effectiveness of counting sort
			
			let wsI = -1, //will be 0 in first round
				currentWS = null,
				sortA = [],
				i=windows.length,
				win, t, j, item;
			
			while(i--) {
				win = windows[i];
				if(win.get_workspace() != currentWS) {
					currentWS = win.get_workspace();
					sortA[++wsI] = [];
				}
				
				sortA[wsI][win._PANEL_DOCKLET_icon._myIndex] = win;
			}
			
			i=sortA.length;
			while(i--) {
				this._appendSeparator();
				t = sortA[i];
				j = t.length;
				while(j--) {
					if(!t[j])
						continue;
					
					win = t[j];
					if(win.has_focus()) {
						item = new St.Label({ text: win.title, style_class: 'focusedOne'});
						this.addActor(item);
					}
					else if(source._myWindow == win) {
						item = new PopupMenu.PopupMenuItem(win.title, {style_class: 'chosenOne'});
						item._myWindow = win;
						this.addMenuItem(item);
					}
					else {
						item = new PopupMenu.PopupMenuItem(win.title, {style_class: 'otherOne'});
						item._myWindow = win;
						this.addMenuItem(item);
					}
				}
			}
		}
    },
	_onActivate: function (actor, child) {
		if (child._myWindow) {
			let win = child._myWindow;
			if(win.get_workspace() != global.screen.get_active_workspace())
				win.get_workspace().activate(global.get_current_time());
			win.activate(global.get_current_time());
		}
		else if (child == this._item_newWin)
			this._source.doNew();
		
		else if (child == this._item_mapWin)
			this._source.doMap();
		
		else if (child == this._item_minimizeWin)
			this._source.doMinimize();
		
		else if (child == this._item_closeWin)
			this._source.doClose();
		else if (actor == this._item_moveRight)
			this._source._swapIcons(true);
		
		else if (actor == this._item_moveLeft)
			this._source._swapIcons(false);
		
		else if(actor == this._item_moveUp && this._source._myWindowClass._myWorkspace.index()) {
			this.close();
			this._source._myWindow.change_workspace_by_index(
					this._source._myWindowClass._myWorkspace.index()-1,
					false, // don't create workspace
					global.get_current_time()
				);
			return;
		}
		else if(actor == this._item_moveDown) {
			this.close();
			this._source._myWindow.change_workspace_by_index(
					this._source._myWindowClass._myWorkspace.index()+1,
					false, // don't create workspace
					global.get_current_time()
				);
			return;
		}
		else if (actor == this._item_quitApp)
			this._source._myApp.request_quit();
		
		else if (actor == this._Item_toggleFav) {
            let favs = AppFavorites.getAppFavorites(),
				favId = this._source._myApp.get_id(),
				isFav = favs.isFavorite(favId);
            if (isFav)
                favs.removeFavorite(favId);
            else
                favs.addFavorite(favId);
        }
		this.close();
		this._removeListener();
	},
	
	_destroy: function() {
		if(this.isOpen)
			this._dock._menuIsOpened = false;
		this._removeListener();
		this.destroy();
	}
}


function SettingsDialog(dock, noWelcome) {
    this._init(dock, noWelcome);
}
SettingsDialog.prototype = {
	__proto__: ModalDialog.ModalDialog.prototype,
	_init: function(dock, noWelcome) {
		this._dock = dock;
		
		ModalDialog.ModalDialog.prototype._init.call(this, { styleClass: "panelDocklet_dialog" });
		
		
		let mainBox = new St.BoxLayout({style_class: 'polkit-dialog-main-layout', vertical: true});
		this.contentLayout.add(mainBox, {x_fill: true, y_fill: true});
			
			this._headline = new St.Label({style_class: 'polkit-dialog-headline'});
			mainBox.add(this._headline, {y_fill: false, y_align: St.Align.START});
			
			this._descripten = new St.Label({style_class: 'polkit-dialog-description'});
			mainBox.add(this._descripten, {y_fill: true});
		
			
			this._messageBox = new St.BoxLayout({style_class: 'polkit-dialog-message-layout', vertical: true});
			mainBox.add(this._messageBox, {y_align: St.Align.START, y_fill: true, x_fill: true});
		
		
		if(noWelcome)
			this._step1();
		else
			this._welcome();
	},
	_welcome: function() {
		this._headline.text = _("Greetings, stranger!");
		this._descripten.text = _("This seems to be the first time, you have enabled the panel docklet. \n\
There are three short steps, that should help you to costumize the new docklet to your needs. \n\
\n\
You can skip this dialog and just use the standard-settings or come back at any time \n\
by clicking with the right mouse-button on the docklet and choosing \"Settings\"\n\n\
\n\
Please feel free to contact me if you find bug, have suggestions, critics or feedback.\n\
It will help, keeping my coding sprit raise up :) \n\
\n\
Have fun!\n\
Jodli (pw3@gmx.at)");
	
		this.setButtons([
				{
					label: _("Cancel"),
					action: Lang.bind(this, function() {
						this.close();
					}),
					key: Clutter.Escape
				},
				{
					label: _("Next"),
					action: Lang.bind(this, function() {
						this._step1();
					})
				}
			]);
	
	},
	_step1: function() {
		this._messageBox.destroy_children(); //FIXME: some objects create NULL-pointer errors in console...
		let item, section;
		
		this._headline.text = _("Appearance (1/3)");
		this._descripten.text = _("The docklet-width is calculated out of the left and right x-position. \n\
The right x-position is subtracted from the center of the screen \n \
\n \
Icon-size also affects the panel height. It dynamically decreases, when the panel-docklet is filled up.");
		
		
		//Icon-size
		section = new St.BoxLayout({vertical: false});
			this._messageBox.add(section);
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Icon-size: ")});
			section.add(item, {expand: true});
		
		item = new St.Entry({style_class: "panelDocklet_preferences_text", text:this._dock._CONF_iconSize.toString()});
			section.add(item, {expand: false});
		item.clutter_text.connect('key-focus-out', Lang.bind(this, function(o) { 
				let s = o.get_text();
				let i = parseInt(o.get_text());
				if(i<10)
					i=10;
				o.set_text(i.toString());
				this._dock._settings.set_int('basic-icon-size', i);
			}));
		
		
		//left x-position
		section = new St.BoxLayout({vertical: false});
			this._messageBox.add(section);
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Left x-position: ")});
			section.add(item, {expand: true});
		
		item = new St.Entry({style_class: "panelDocklet_preferences_text", text:this._dock._CONF_dockX_left.toString()});
			section.add(item, {expand: false});
		item.clutter_text.connect('key-focus-out', Lang.bind(this, function(o) {
				this._dock._settings.set_int('dock-x-left', parseInt(o.get_text()));
			}));
		
		
		//right x-position
		section = new St.BoxLayout({vertical: false});
			this._messageBox.add(section);
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Right x-position (subtracted from screen-center): ")});
			section.add(item, {expand: true});
		
		item = new St.Entry({style_class: "panelDocklet_preferences_text", text:this._dock._CONF_dockX_right.toString()});
			section.add(item, {expand: false});
		item.clutter_text.connect('key-focus-out', Lang.bind(this, function(o) {
				this._dock._settings.set_int('dock-x-right-from-center', parseInt(o.get_text()));
			}));
			
		
		this.setButtons([
			{
				label: _("Close"),
				action: Lang.bind(this, function() {
					this.close();
				}),
				key: Clutter.Escape
			},
			{
				label: _("Next"),
				action: Lang.bind(this, function() {
					
					this._step2();
				}),
			}
		]);
	},
	_step2: function() {
		let section, group, button, item, combo;
		
		this._messageBox.destroy_children(); //FIXME: some objects create NULL-pointer errors in console...
		this._headline.text = _("Features (2/3)");
		this._descripten.text = _("Hthis features affect the appereance and behaviour of the docklet");
		
		section = new PopupMenu.PopupComboMenu();
			this._messageBox.add(section.actor);
		
		
		//middle click
		group = new St.BoxLayout({vertical: false, style:"padding: 0 5px"});
			section.addActor(group, {x_fill: true});
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Action for middleclick on Docklet-icons: ")});
			group.add(item, {expand:true});
		
		combo = new PopupMenu.PopupComboBoxMenuItem({ style_class: 'panelDocklet_combo' });
			group.add(combo.actor);
		
		item = new PopupMenu.PopupMenuItem(_("Open new window"));
			combo.addMenuItem(item, NEW_WINDOW);
		item = new PopupMenu.PopupMenuItem(_("Minimize window"));
			combo.addMenuItem(item, MINIMIZE_WINDOW);
		item = new PopupMenu.PopupMenuItem(_("Close window"));
			combo.addMenuItem(item, CLOSE_WINDOW);
		
		combo.setActiveItem(this._dock._settings.get_enum("middle-click-action"));
		combo.connect('active-item-changed', Lang.bind(this, function(menuItem, id) {
				this._dock._settings.set_enum("middle-click-action", id);
			}));
		
		
		
		//Auto-hide
		group = new St.BoxLayout({vertical: false, style:"padding: 0 5px"});
			section.addActor(group, {x_fill: true});
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Auto-hide of docklet:")});
			group.add(item, {expand:true});
		
		button = new St.Button({reactive: true, can_focus: false});
			group.add(button);
		this._autohide_Switch = new PopupMenu.Switch(this._dock._CONF_autohide);
			button.set_child(this._autohide_Switch.actor);
		
		button.connect("button-press-event",  Lang.bind(this, function() {
				this._autohide_Switch.toggle();
				this._dock._settings.set_boolean("autohide", this._autohide_Switch.state);
			}));
		
		
		
		//list-workspaces-on-open
		group = new St.BoxLayout({vertical: false, style:"padding: 0 5px"});
			section.addActor(group, {x_fill: true});
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("List other workspaces:")});
			group.add(item, {expand:true});
		
		button = new St.Button({reactive: true, can_focus: false});
			group.add(button);
		
		this.showAllWs_Switch = new PopupMenu.Switch(this._dock._CONF_showAllWS);
			button.set_child(this.showAllWs_Switch.actor);
		
		button.connect("button-press-event",  Lang.bind(this, function() {
				this.showAllWs_Switch.toggle();
				this._dock._settings.set_boolean("list-workspaces-on-open", this.showAllWs_Switch.state);
			}));
			
			
		//show-workspace-number
		group = new St.BoxLayout({vertical: false, style:"padding: 0 5px"});
			section.addActor(group, {x_fill: true});
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Show workspace-indicator:")});
			group.add(item, {expand:true});
		
		button = new St.Button({reactive: true, can_focus: false});
			group.add(button);
		
		this.showWsLine_Switch = new PopupMenu.Switch(this._dock._CONF_showWSline);
			button.set_child(this.showWsLine_Switch.actor);
		
		button.connect("button-press-event",  Lang.bind(this, function() {
				this.showWsLine_Switch.toggle();
				this._dock._settings.set_boolean("show-workspace-number", this.showWsLine_Switch.state);
			}));
		
		
		
		this.setButtons([
			{
				label: _("back"),
				action: Lang.bind(this, function() {
					this._step1();
				}),
				key: Clutter.Escape
			},
			{
				label: _("Close"),
				action: Lang.bind(this, function() {
					this.close();
				}),
				key: Clutter.Escape
			},
			{
				label: _("Next"),
				action: Lang.bind(this, function() {
					this._step3();
				})
			}
		]);
		
	},
	_step3: function() {
		this._messageBox.destroy_children();//FIXME: some objects create NULL-pointer errors in console...
		let item, button, section;
		
		this._headline.text = _("Auto-hide (3/3)");
		this._descripten.text = _("This settings are only used, if auto-hide is enabled");
		
		
		//hide to
		section = new St.BoxLayout({vertical: false});
			this._messageBox.add(section);
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Hide docklet-bottom to position: ")});
			section.add(item, {expand: true});
		
		item = new St.Entry({style_class: "panelDocklet_preferences_text", text:this._dock._CONF_hideToBottomPos.toString()});
			section.add(item, {expand: false});
		item.clutter_text.connect('key-focus-out', Lang.bind(this, function(o) {
				this._dock._settings.set_int('bottom-hide-to', parseInt(o.get_text()));
			}));
		
		//time-until-hide
		section = new St.BoxLayout({vertical: false});
			this._messageBox.add(section);
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Waiting-time until docklet starts hiding: ")});
			section.add(item, {expand: true});
		
		item = new St.Entry({style_class: "panelDocklet_preferences_text", text:this._dock._CONF_menuTimeout.toString()});
			section.add(item, {expand: false});
		item.clutter_text.connect('key-focus-out', Lang.bind(this, function(o) {
				this._dock._settings.set_int('time-until-hide', parseInt(o.get_text()));
			}));
		
		//hide-time
		section = new St.BoxLayout({vertical: false});
			this._messageBox.add(section);
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Hiding-duration: ")});
			section.add(item, {expand: true});
		
		item = new St.Entry({style_class: "panelDocklet_preferences_text", text:this._dock._CONF_hideTime.toString()});
			section.add(item, {expand: false});
		item.clutter_text.connect('key-focus-out', Lang.bind(this, function(o) {
				this._dock._settings.set_int('hide-time', parseInt(o.get_text()));
			}));
		
		
		//show-time
		section = new St.BoxLayout({vertical: false});
			this._messageBox.add(section);
		
		item = new St.Label({style_class: 'polkit-dialog-description', text: _("Showing-duration: ")});
			section.add(item, {expand: true});
		
		item = new St.Entry({style_class: "panelDocklet_preferences_text", text:this._dock._CONF_showTime.toString()});
			section.add(item, {expand: false});
		item.clutter_text.connect('key-focus-out', Lang.bind(this, function(o) {
				this._dock._settings.set_int('show-time', parseInt(o.get_text()));
			}));
		
		
		
		this._messageBox.add(section);
		
		this.setButtons([
			{
				label: _("back"),
				action: Lang.bind(this, function() {
					this._step2();
				}),
				key: Clutter.Escape
			},
			{
				label: _("Close"),
				action: Lang.bind(this, function() {
					this.close();
				})
			}
		]);
	}
}


function initTranslations(metadata) {
	//let localeDir = metadata.dir.get_child('locale').get_path();
	let localeDir = metadata.localedir;
	imports.gettext.bindtextdomain('gnome-shell-extensions', localeDir);
}

function getSettings(metadata, extension_id) {
	//let schemaDir = metadata.dir.get_child('schemas').get_path();
	let schemaDir = metadata.localedir;
	let schemaSource = Gio.SettingsSchemaSource.new_from_directory(
			schemaDir,
			Gio.SettingsSchemaSource.get_default(),
			false
		);
	let schema = schemaSource.lookup('org.gnome.shell.extensions.' + extension_id, false);
	return new Gio.Settings({ settings_schema: schema });
}




function init(metadata) {
	extention_path = metadata.path;
	//imports.gettext.bindtextdomain('gnome-shell-extensions', extensionMeta.localedir); //blind copy...
}

function preferences(dock, noWelcome) {
	if(noWelcome)
		new SettingsDialog(dock, true).open();
	else
		Mainloop.timeout_add(2000, Lang.bind(this, function() {
			dock._settings.set_boolean("first-time", false);
			new SettingsDialog(dock).open();
		}));
}
function reload() {
	disable();
	enable();
}
function enable() {
	myPanelDocklet=new panelDocklet();
}
function disable() {
	myPanelDocklet._destroy();
}

/*
gnome-shell --replace

*/
